#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"

// -1 is used in many places to designate
// the absence of an optional object stub
#ifndef UNASSIGNED
	#define UNASSIGNED  0xFFFF
#endif

//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

void botInfo::WriteIni()
{
	char Path[512]; //file path
	GetCurrentDirectory(512, Path); //grab the path
	strcat(Path, "\\FlagBot.ini"); //append the file name
				
	/**********AntiWarp***********/
	WritePrivateProfileString("AntiWarp", "AWCheck", getString(enabled, 2, 0, true), Path); 
	WritePrivateProfileString("AntiWarp", "Radius", getString(radius, 10, 0, true), Path);
	WritePrivateProfileString("AntiWarp", "Delay", getString(delay, 10, 0, true), Path);
	WritePrivateProfileString("AntiWarp", "Warn",getString(warn, 10, 0, true), Path);
					
	/**********Flagging***********/
	WritePrivateProfileString("Flagging", "Jperiodic",getString(jperiodic, 2, 0, true), Path);
	WritePrivateProfileString("Flagging", "Periodic", getString(periodic, 10, 0, true), Path);
	WritePrivateProfileString("Flagging", "Jgreet", getString(jgreet, 2, 0, true), Path);
	WritePrivateProfileString("Flagging", "Victory", getString(victory, 2, 0, true), Path);
	WritePrivateProfileString("Flagging", "FlagTime", getString(flagtime, 10, 0, true), Path);
	WritePrivateProfileString("Flagging", "SpamFlags", getString(spamflags, 2, 0, true), Path);

	/**********Offenses***********/
	WritePrivateProfileString("Offenses", "TeamKill", getString(teamkill, 2, 0, true), Path);
	WritePrivateProfileString("Offenses", "SafeKill", getString(safekill, 2, 0, true), Path);
	WritePrivateProfileString("Offenses", "SKRadius", getString(skradius, 2, 0, true), Path);
	WritePrivateProfileString("Offenses", "Warnings", getString(warnings, 10, 0, true), Path);
	WritePrivateProfileString("Offenses", "Grace", getString(grace, 10, 0, true), Path);
	WritePrivateProfileString("Offenses", "BanTime", getString(bantime, 10, 0, true), Path);

	/**********Teams***********/
	WritePrivateProfileString("Teams", "CheckEvenTeams", getString(checkeventeams, 2, 0, true), Path);
	WritePrivateProfileString("Teams", "UnevenPlayers", getString(unevenplayers, 10, 0, true), Path);
	WritePrivateProfileString("Teams", "UnevenTeams", getString(uneventeams, 10, 0, true), Path);
}

Sint32 getNumeric(Sint32 y)
{
	return (y * 20 / TILE_MAX_Y) + 1;
}

char getAlpha(Sint32 x)
{
	return char('A' + x * 20 / TILE_MAX_X);
}

String getCoords(Sint32 x, Sint32 y)
{
	String s;

	char alpha[2];
	alpha[0] = getAlpha(x);
	alpha[1] = '\0';

	s += alpha;
	s += getNumeric(y);

	return s;
}

int botInfo::getTeamPlayers(int team)
{
	_listnode <Player> *parse = playerlist->head;
	int count = 0;

	while (parse)
	{
		Player *p = parse->item;

		if (p->ship != SHIP_Spectator)
		if (p->team == team)
			++count;

		parse = parse->next;
	}

	return count;
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];

			if (countdown[0] == 0)
			{
				if (jperiodic && new_jackpot && (jackpot > 0))
				{
					new_jackpot = false;
					sendPublic(SND_BassBeep, "*arena Jackpot just passed " + String(jackpot) + " points.");
				}

				countdown[0] = periodic * 60;
			}

			if (countdown[1] == 0)
			{
				countdown[1] = 10;

				sendPublic("?jackpot");
			}

			if (countdown[2] == 0)
			{
				_listnode <Player> *parse = playerlist->head;

				while (parse)
				{
					Player *p = parse->item;

					int count = get_tag(p, TAG_OFFENSES);
					if (count > 0)
					{
						set_tag(p, TAG_OFFENSES, count - 1);
					}

					parse = parse->next;
				}

				countdown[2] = grace * 60;
			}

			if (countdown[3] == 0)
			{
				if (spamflags)
				{
					_listnode <Flag> *parse;
					String last_coords;
					int count;
					String output;
					bool seen;

					parse = flaglist->head;
					last_coords = "";
					count = 0;
					seen = false;

					while (parse)
					{
						Flag *f = parse->item;
						parse = parse->next;

						if (f->team == UNASSIGNED) continue;

						String coords = getCoords(f->x, f->y);

						if (coords != last_coords)
						{
							if (count > 0)
							{
								if (seen)
									output += ", ";
								else
									seen = true;

								if (count > 1)
								{
									output += count;
									output += " at ";
								}
								output += last_coords;

								count = 0;
							}
						}

						++count;

						last_coords = coords;
					}

					if (count > 0)
					{
						if (seen)
							output += ", ";

						if (count > 1)
						{
							output += count;
							output += " at ";
						}
						output += last_coords;

						count = 0;
					}

					if (!output.IsEmpty())
						sendPublic("*arena Owned flags: " + output);


					parse = flaglist->head;
					last_coords = "";
					output = "";
					count = 0;
					seen = false;

					while (parse)
					{
						Flag *f = parse->item;
						parse = parse->next;

						if (f->team != UNASSIGNED) continue;

						String coords = getCoords(f->x, f->y);

						if (coords != last_coords)
						{
							if (count > 0)
							{
								if (seen)
									output += ", ";
								else
									seen = true;

								if (count > 1)
								{
									output += count;
									output += " at ";
								}
								output += last_coords;

								count = 0;
							}
						}

						++count;

						last_coords = coords;
					}

					if (count > 0)
					{
						if (seen)
							output += ", ";

						if (count > 1)
						{
							output += count;
							output += " at ";
						}
						output += last_coords;

						count = 0;
					}

					if (!output.IsEmpty())
						sendPublic("*arena Neuted flags: " + output);
				}

				countdown[3] = flagtime * 60;
			}
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
			bool biller_online = *(bool*)&event.p[2];
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
			killTags();
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];

			set_tag(p, TAG_STAT_FS, get_tag(p, TAG_STAT_FS) + 1);
			set_tag(p, TAG_FLAGTIMER, GetTickCount());
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];

			if (p->flagCount > 0)
			{
				set_tag(p, TAG_STAT_FD, get_tag(p, TAG_STAT_FD) + p->flagCount);

				int ticks = get_tag(p, TAG_FLAGTIMER);
				if (ticks > 0)
				{
					set_tag(p, TAG_STAT_FT, get_tag(p, TAG_STAT_FT) + (GetTickCount() - ticks));
					set_tag(p, TAG_FLAGTIMER, 0);
				}
			}
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];

			if (gametime == 0)
			{	// game reset
				gametime = GetTickCount();

				_listnode <Player> *parse = playerlist->head;

				while (parse)
				{
					Player *p = parse->item;

					set_tag(p, TAG_STAT_W, 0);
					set_tag(p, TAG_STAT_L, 0);
					set_tag(p, TAG_STAT_TK, 0);
					set_tag(p, TAG_STAT_G, 0);
					set_tag(p, TAG_STAT_FL, 0);
					set_tag(p, TAG_STAT_FK, 0);
					set_tag(p, TAG_STAT_FS, 0);
					set_tag(p, TAG_STAT_FD, 0);
					set_tag(p, TAG_STAT_FT, 0);
					set_tag(p, TAG_STAT_KP, 0);
					set_tag(p, TAG_FLAGTIMER, 0);

					parse = parse->next;
				}
			}
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];

			if (!victory)
			{
				gametime = 0;
				break;
			}

			/*
				Winners jackpot
				Game time

				Name
				Wins
				Losses
				Team kills
				Goals
				Rating
				Flags lost by deaths
				Flags gotten from kills
				Flags scavenged
				Flags dropped
				Flag time
				Kill points
			*/

			String names[NUM_STATS];
			int best[NUM_STATS];

			for (int i = 0; i < NUM_STATS; ++i)
				best[i] = 0;

			_listnode <Player> *parse;

			String str[NUM_STATS];
			Uint32 len[NUM_STATS];

			for (i = 0; i < NUM_STATS; ++i)
				len[i] = 0;

			parse = playerlist->head;
			while (parse)
			{
				Player *p = parse->item;

				if (p->team == team)
				{
					int stat[NUM_STATS];

					stat[1] = get_tag(p, TAG_STAT_W);
					stat[2] = get_tag(p, TAG_STAT_L);
					stat[3] = get_tag(p, TAG_STAT_TK);
					stat[4] = get_tag(p, TAG_STAT_G);
					stat[6] = get_tag(p, TAG_STAT_FL);
					stat[7] = get_tag(p, TAG_STAT_FK);
					stat[8] = get_tag(p, TAG_STAT_FS);
					stat[9] = get_tag(p, TAG_STAT_FD);
					stat[10]= get_tag(p, TAG_STAT_FT) / 1000;
					stat[11]= get_tag(p, TAG_STAT_KP);

					stat[5] = 1000 * (stat[11] + (stat[1] - stat[2]) * 10) / (stat[1] + 100);
					if (stat[5] < 0) stat[5] = 0;

					str[0] = p->name;
					for (i = 1; i < NUM_STATS; ++i)
					{
						if (i == 5)
						{	// ratings look like floating-point numbers
							int n, f;
							n = stat[5] / 100;
							f = stat[5] % 100;

							str[5] = n;
							str[5] += ".";
							str[5] += getString(f, 10, 2, false);
						}
						else
						{
							str[i] = stat[i];
						}
					}

					for (i = 0; i < NUM_STATS; ++i)
						if (len[i] < str[i].len)
							len[i] = str[i].len;
				}

				parse = parse->next;
			}

			len[0] = 20;

			str[0] = "Name";
			str[1] = "W";
			str[2] = "L";
			str[3] = "TK";
			str[4] = "G";
			str[5] = "Rating";
			str[6] = "FL";
			str[7] = "FK";
			str[8] = "FS";
			str[9] = "FD";
			str[10] = "FT";
			str[11] = "KP";

			for (i = 0; i < NUM_STATS; ++i)
			{
				if (len[i] < str[i].len)
					len[i] = str[i].len;

				str[i] = str[i].lfill(len[i]);
			}

			String bigline;
			bigline += str[0];
			for (i = 1; i < NUM_STATS; ++i)
			{
				bigline += "  ";
				bigline += str[i];
			}

			char *__bars = "--------------------------------------------------------------------------------------------------------";
			String bars(__bars, bigline.len);

			String jp;
			jp += "Winners Jackpot(";
			jp += reward;
			jp += ")";

			gametime = (GetTickCount() - gametime) / 1000;

			String gt;
			gt += "Game Time:";
			gt += getString(gametime / 3600, 10, 2, true);
			gt += "h ";
			gt += getString((gametime / 60) % 60, 10, 2, true);
			gt += "m ";
			gt += getString(gametime % 60, 10, 2, true);
			gt += "s";

			sendPublic("*arena |" + bars + "|");

			sendPublic("*arena |" + jp + gt.lfill(bigline.len - jp.len) + "|");

			sendPublic("*arena |" + bigline + "|");

			sendPublic("*arena |" + bars + "|");

			parse = playerlist->head;
			while (parse)
			{
				Player *p = parse->item;

				if (p->team == team)
				{
					int stat[NUM_STATS];
					stat[1] = get_tag(p, TAG_STAT_W);
					stat[2] = get_tag(p, TAG_STAT_L);
					stat[3] = get_tag(p, TAG_STAT_TK);
					stat[4] = get_tag(p, TAG_STAT_G);
					stat[6] = get_tag(p, TAG_STAT_FL);
					stat[7] = get_tag(p, TAG_STAT_FK);
					stat[8] = get_tag(p, TAG_STAT_FS);
					stat[9] = get_tag(p, TAG_STAT_FD);
					stat[10]= get_tag(p, TAG_STAT_FT) / 1000;
					stat[11]= get_tag(p, TAG_STAT_KP);

					stat[5] = 1000 * (stat[11] + (stat[1] - stat[2]) * 10) / (stat[1] + 100);
					if (stat[5] < 0) stat[5] = 0;

					str[0] = p->name;
					for (int i = 1; i < NUM_STATS; ++i)
					{
						if (i == 5)
						{	// ratings look like floating-point numbers
							int n, f;
							n = stat[5] / 100;
							f = stat[5] % 100;

							str[5] = n;
							str[5] += ".";
							str[5] += getString(f, 10, 2, false);
						}
						else
						{
							str[i] = stat[i];
						}
					}

					for (i = 0; i < NUM_STATS; ++i)
						str[i] = str[i].lfill(len[i]);

					String s = "|";
					s += str[0];
					for ( i = 1; i < NUM_STATS; ++i)
					{
						s += "  ";
						s += str[i];
					}
					s += "|";
					sendPublic("*arena " + s);
				}

				parse = parse->next;
			}

			sendPublic("*arena |" + bars + "|");

			parse = playerlist->head;

			while (parse)
			{
				Player *p = parse->item;

				int stat[NUM_STATS];
				stat[1] = get_tag(p, TAG_STAT_W);
				stat[2] = get_tag(p, TAG_STAT_L);
				stat[3] = get_tag(p, TAG_STAT_TK);
				stat[4] = get_tag(p, TAG_STAT_G);
				stat[6] = get_tag(p, TAG_STAT_FL);
				stat[7] = get_tag(p, TAG_STAT_FK);
				stat[8] = get_tag(p, TAG_STAT_FS);
				stat[9] = get_tag(p, TAG_STAT_FD);
				stat[10]= get_tag(p, TAG_STAT_FT) / 1000;
				stat[11]= get_tag(p, TAG_STAT_KP);
				stat[5] = 1000 * (stat[11] + (stat[1] - stat[2]) * 10) / (stat[1] + 100);

				for (int i = 1; i < NUM_STATS; ++i)
				{
					if (stat[i] > best[i])
					{
						names[i] = p->name;
						best[i] = stat[i];
					}
					else if (stat[i] == best[i])
					{
						names[i] += ", ";
						names[i] += p->name;
					}
				}

				parse = parse->next;
			}

			if (best[5] > 0)	sendPublic("*arena [Rg] Highest rating: " + names[5]);
			if (best[1] > 0)	sendPublic("*arena [W ] Most deadly: " + names[1] + " (" + String(best[1]) + " wins)");
			if (best[2] > 0)	sendPublic("*arena [L ] Most deaths: " + names[2] + " (" + String(best[2]) + " losses)");
			if (best[3] > 0)	sendPublic("*arena [TK] Biggest pain in the neck: " + names[3] + " (" + String(best[3]) + " TK's)");
			if (best[4] > 0)	sendPublic("*arena [G ] Most goals: " + names[4] + " (" + String(best[4]) + " goals)");
			if (best[6] > 0)	sendPublic("*arena [FL] Most flags lost by deaths: " + names[6] + " (" + String(best[6]) + " flags)");
			if (best[7] > 0)	sendPublic("*arena [FK] Most flags acquired by kills: " + names[7] + " (" + String(best[7]) + " flags)");
			if (best[8] > 0)	sendPublic("*arena [FS] Most flags scavenged: " + names[8] + " (" + String(best[8]) + " flags)");
			if (best[9] > 0)	sendPublic("*arena [FD] Most flags successfully dropped: " + names[9] + " (" + String(best[9]) + " flags)");
			if (best[10] > 0)	sendPublic("*arena [FT] Most flag time: " + names[10] + " (" + String(best[10]) + " seconds)");
			if (best[11] > 0)	sendPublic("*arena [KP] Most points from kills: " + names[11] + " (" + String(best[11]) + " points)");

			gametime = 0;
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_BallMove:
		{
			PBall *ball = (PBall*)event.p[0];

			int ident = ball->carrier;

			if (ident == UNASSIGNED)
			{	// dropped
			}
			else
			{	// carried
				Player *p = getPlayer(ident);
			}
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];

			if (jgreet)
			if (jackpot > 0)
			{
				sendPrivate(p, "Jackpot is currently " + String(jackpot));
			}
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];

			int ticks = GetTickCount();

			if ((p->awarp) && (ticks - get_tag(p, TAG_TIMER) >= 1000) &&
				(p->tile.x >= 512 - radius) && (p->tile.y >= 512 - radius) &&
				(p->tile.x <= 512 + radius) && (p->tile.y <= 512 + radius))
			{
				if (delay > 0)
				{
					if (get_tag(p, TAG_DELAY) == 0)
					{
						set_tag(p, TAG_DELAY, ticks);
						break;
					}
					else if (ticks - get_tag(p, TAG_DELAY) < delay)
					{
						break;
					}
				}

				set_tag(p, TAG_DELAY, 0);

				sendPrivate(p, "*prize #-20");
				set_tag(p, TAG_TIMER, ticks);

				if ((warn == 2) || ((warn == 1) && (get_tag(p, TAG_WARNED) == false)))
				{
					sendPrivate(p, "*warn No anti-warp in the center. Yoink!");
					set_tag(p, TAG_WARNED, true);
				}
			}
			else
			{
				set_tag(p, TAG_DELAY, 0);
			}
		}
		break;
	case EVENT_PlayerWeapon:
		{
			Player *p = (Player*)event.p[0];
			weaponInfo wi;
			wi.n = *(Uint16*)&event.p[1];
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];
			int bounty = *(int *)&event.p[2];

			set_tag(k, TAG_STAT_W, get_tag(k, TAG_STAT_W) + 1);
			set_tag(k, TAG_STAT_KP, get_tag(k, TAG_STAT_KP) + bounty);
			set_tag(p, TAG_STAT_L, get_tag(p, TAG_STAT_L) + 1);

			if (p->flagCount > 0)
			{
				set_tag(p, TAG_STAT_FL, get_tag(p, TAG_STAT_FL) + p->flagCount);
				set_tag(k, TAG_STAT_FK, get_tag(k, TAG_STAT_FK) + p->flagCount);

				int ticks = get_tag(p, TAG_FLAGTIMER);
				if (ticks > 0)
				{
					set_tag(p, TAG_STAT_FT, get_tag(p, TAG_STAT_FT) + (GetTickCount() - ticks));
					set_tag(p, TAG_FLAGTIMER, 0);
				}
			}

			if (p->team == k->team)
			{
				set_tag(k, TAG_STAT_TK, get_tag(k, TAG_STAT_TK) + 1);

				int count = get_tag(k, TAG_OFFENSES);
				set_tag(k, TAG_OFFENSES, count + 1);

				if (teamkill)
				{
					if (count >= warnings)
					{
						sendPrivate(k, "*kill " + String(bantime));
					}
					else
					{
						sendPrivate(k, "*warn No teamkilling in this arena. You have been warned");
					}
				}
			}
			else if (p->safety)
			{
				if (safekill)
				if (!p->turret)
				if (!skradius ||
					((p->tile.x >= 512 - radius) && (p->tile.y >= 512 - radius) &&
					(p->tile.x <= 512 + radius) && (p->tile.y <= 512 + radius)))
				{
					int count = get_tag(k, TAG_OFFENSES);
					set_tag(k, TAG_OFFENSES, count + 1);

					if (count >= warnings)
					{
						sendPrivate(k, "*kill " + String(bantime));
					}
					else
					{
						sendPrivate(k, "*warn No safekilling in this arena. You have been warned");
					}
				}
			}

			set_tag(p, TAG_TIMER, 0);
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
			int prize = *(int*)&event.p[1];

			if (prize == 20)
			{
				set_tag(p, TAG_TIMER, 0);
			}
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldship = (Uint16)event.p[1];
			Uint16 oldteam = (Uint16)event.p[2];

			set_tag(p, TAG_TIMER, 0);

			if (p->flagCount > 0)
			{
				int ticks = get_tag(p, TAG_FLAGTIMER);
				if (ticks > 0)
				{
					set_tag(p, TAG_STAT_FT, get_tag(p, TAG_STAT_FT) + (GetTickCount() - ticks));
					set_tag(p, TAG_FLAGTIMER, 0);
				}
			}
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];
			Uint16 oldship = (Uint16)event.p[2];

			set_tag(p, TAG_TIMER, 0);

			if (p->flagCount > 0)
			{
				int ticks = get_tag(p, TAG_FLAGTIMER);
				if (ticks > 0)
				{
					set_tag(p, TAG_STAT_FT, get_tag(p, TAG_STAT_FT) + (GetTickCount() - ticks));
					set_tag(p, TAG_FLAGTIMER, 0);
				}
			}

			if (checkeventeams)
			if (oldship != SHIP_Spectator)
			{
				int teams[MAX_TEAMS];
				int low = 1000, high = 0;
				String counts;

				for (int i = 0; i < uneventeams; ++i)
				{
	                teams[i] = getTeamPlayers(i);

					if (teams[i] > high)	high = teams[i];
					if (teams[i] < low)		low = teams[i];

					counts += "  ";
					counts += i;
					counts += ":(";
					counts += teams[i];
					counts += ")";
				}

				if (high - low >= unevenplayers)
				{
					sendPublic("*arena Teams are uneven." + counts);
				}
			}
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];
			Uint16 oldship = (Uint16)event.p[2];

			set_tag(p, TAG_TIMER, 0);

			if (p->flagCount > 0)
			{
				int ticks = get_tag(p, TAG_FLAGTIMER);
				if (ticks > 0)
				{
					set_tag(p, TAG_STAT_FT, get_tag(p, TAG_STAT_FT) + (GetTickCount() - ticks));
					set_tag(p, TAG_FLAGTIMER, 0);
				}
			}

			if (checkeventeams)
			if (oldship != SHIP_Spectator)
			{
				int teams[MAX_TEAMS];
				int low = 1000, high = 0;
				String counts;

				for (int i = 0; i < uneventeams; ++i)
				{
	                teams[i] = getTeamPlayers(i);

					if (teams[i] > high)	high = teams[i];
					if (teams[i] < low)		low = teams[i];

					counts += "  ";
					counts += i;
					counts += ":(";
					counts += teams[i];
					counts += ")";
				}

				if (high - low >= unevenplayers)
				{
					sendPublic("*arena Teams are uneven." + counts);
				}
			}
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];

			if (checkeventeams)
			if (p->ship != SHIP_Spectator)
			{
				int teams[MAX_TEAMS];
				int low = 1000, high = 0;
				String counts;

				for (int i = 0; i < uneventeams; ++i)
				{
	                teams[i] = getTeamPlayers(i);

					if (p->team == i) --teams[i];

					if (teams[i] > high)	high = teams[i];
					if (teams[i] < low)		low = teams[i];

					counts += "  ";
					counts += i;
					counts += ":(";
					counts += teams[i];
					counts += ")";
				}

				if (high - low >= unevenplayers)
				{
					sendPublic("*arena Teams are uneven." + counts);
				}
			}

			killTags(p);
		}
		break;
//////// Turret ////////
	case EVENT_SelfShipReset:
		{
		}
		break;
	case EVENT_SelfPrize:
		{
			Uint16 prize = (Uint16)event.p[0];
			Uint16 count = (Uint16)event.p[1];
		}
		break;
	case EVENT_SelfUFO:
		{
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
					if (sound == SND_None)
					if (CMPSTART("Current Jackpot: ", msg))
					{
						int njp = getInteger(msg + 17, 10);

						if (njp != jackpot)
						{
							new_jackpot = true;
							jackpot = njp;
						}
					}

					if (sound == SND_Goal)
					if (CMPSTART("Enemy Goal! by ", msg))
					{
						// A:Enemy Goal! by catid  Reward:25000

						for (int i = STRLEN(msg) - 1; i >= 0; --i)
							if (msg[i] == ' ')
								break;
						msg[i - 1] = '\0';

						char *name = msg + 15;

						_listnode <Player> *parse = playerlist->head;
						while (parse)
						{
							Player *p = parse->item;

							if (CMPSTR(name, p->name))
							{
								set_tag(p, TAG_STAT_G, get_tag(p, TAG_STAT_G) + 1);

								break;
							}

							parse = parse->next;
						}
					}

					if (sound == SND_VictoryBell)
					if (CMPSTR("Flag game reset.", msg))
					{
						gametime = GetTickCount();

						_listnode <Player> *parse = playerlist->head;

						while (parse)
						{
							Player *p = parse->item;

							set_tag(p, TAG_STAT_W, 0);
							set_tag(p, TAG_STAT_L, 0);
							set_tag(p, TAG_STAT_TK, 0);
							set_tag(p, TAG_STAT_G, 0);
							set_tag(p, TAG_STAT_FL, 0);
							set_tag(p, TAG_STAT_FK, 0);
							set_tag(p, TAG_STAT_FS, 0);
							set_tag(p, TAG_STAT_FD, 0);
							set_tag(p, TAG_STAT_FT, 0);
							set_tag(p, TAG_STAT_KP, 0);
							set_tag(p, TAG_FLAGTIMER, 0);

							parse = parse->next;
						}
					}
				}
				break;
			case MSG_PublicMacro:		if (!p) break;
				{
				}
				break;
			case MSG_Public:			if (!p) break;
				{
				}
				break;
			case MSG_Team:				if (!p) break;
				{
				}
				break;
			case MSG_TeamPrivate:		if (!p) break;
				{
				}
				break;
			case MSG_Private:			if (!p) break;
				{
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
	case EVENT_PositionHook:
		{
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major = HIWORD(*(int*)&event.p[0]);
			int minor = LOWORD(*(int*)&event.p[0]);
			callback = (CALL_COMMAND)event.p[1];
			playerlist = (CALL_PLIST)event.p[2];
			flaglist = (CALL_FLIST)event.p[3];
			map = (CALL_MAP)event.p[4];

			if (major > CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin is out of date."));

				CONNECTION_DENIED = true;

				return;
			}
			else if ((major < CORE_MAJOR_VERSION) || (minor < CORE_MINOR_VERSION))
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin requires the latest version of MERVBot."));

				CONNECTION_DENIED = true;

				return;
			}
			else
			{	
				tell(makeEcho("DLL plugin connected."));
				
				char buffer[6]; //holds the input from the file
				char Path[512]; //file path
				GetCurrentDirectory(512, Path); //grab the path
				strcat(Path, "\\FlagBot.ini"); //append the file name
				
				/**********AntiWarp***********/
				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("AntiWarp", "AWCheck", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{enabled=true;}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("AntiWarp", "Radius", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					radius= getInteger(buffer,10);
					if(radius < 0 || radius > 512)
					{radius=25;}
				}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("AntiWarp", "Delay", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					delay= getInteger(buffer,10);
					if(delay < 0 || delay > 5000)
					{delay=0;}
				}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("AntiWarp", "Warn", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					warn= getInteger(buffer,10);
					if(warn < 0 || warn > 2)
					{warn=1;}
				}
				
				/**********Flagging***********/
				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Flagging", "Jperiodic", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{jperiodic=true;}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Flagging", "Periodic", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					periodic= getInteger(buffer,10);
					if(periodic < 1 || periodic > 60)
					{periodic=10;}
				}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Flagging", "Jgreet", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{jgreet=true;}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Flagging", "Victory", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{victory=true;}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Flagging", "FlagTime", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					flagtime= getInteger(buffer,10);
					if(flagtime < 1 || flagtime > 60)
					{flagtime=5;}
				}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Flagging", "SpamFlags", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{spamflags=true;}

				/**********Offenses***********/
				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Offenses", "TeamKill", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{teamkill=true;}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Offenses", "SafeKill", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{safekill=true;}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Offenses", "SKRadius", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{skradius=true;}
				
				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Offenses", "Warnings", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					warnings= getInteger(buffer,10);
					if(warnings < 1 || warnings > 10)
					{warnings=2;}
				}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Offenses", "Grace", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					grace= getInteger(buffer,10);
					if(grace < 1 || grace > 14440)
					{grace=5;}
				}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Offenses", "BanTime", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					bantime= getInteger(buffer,10);
					if(bantime < 1 || bantime > 14440)
					{bantime=10;}
				}

				/**********Teams***********/
				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Teams", "CheckEvenTeams", NULL, buffer, 2, Path);
				if(buffer[0] && isNumeric(buffer) && getInteger(buffer, 10) > 0)
				{checkeventeams=true;}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Teams", "UnevenPlayers", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					unevenplayers= getInteger(buffer,10);
					if(unevenplayers < 2 || unevenplayers > 10)
					{unevenplayers=3;}
				}

				ZeroMemory(buffer, 6); 
				GetPrivateProfileString("Teams", "UnevenTeams", NULL, buffer, 5, Path);
				if(buffer[0] && isNumeric(buffer))
				{
					uneventeams= getInteger(buffer,10);
					if(uneventeams < 2 || uneventeams >10)
					{uneventeams=2;}
				}
			}
		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	};
}

Player *botInfo::getPlayer(int ident)
{
	_listnode <Player> *parse = playerlist->head;

	while (parse)
	{
		Player *p = parse->item;

		if (p->ident == ident)
		{
			return p;
		}

		parse = parse->next;
	}

	return NULL;
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Tag emulation ////////

int botInfo::get_tag(Player *p, int index)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
			return tag->data;

		parse = parse->next;
	}

	return 0;
}

void botInfo::set_tag(Player *p, int index, int data)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
		{
			tag->data = data;
			return;
		}

		parse = parse->next;
	}

	tag = new PlayerTag;
	tag->p = p;
	tag->index = index;
	tag->data = data;
	taglist.append(tag);
}

void botInfo::killTags(Player *p)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;
		parse = parse->next;

		if (tag->p != p) continue;

		taglist.kill(tag);
	}
}

void botInfo::killTags()
{
	taglist.clear();
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
